@@ -24,14 +24,9 @@
  * Use is subject to license terms.
  */
 
-#pragma ident	"@(#)tst.ret1.c	1.1	06/08/28 SMI"
-
 #include <signal.h>
-#if !defined(__APPLE__)
-#include <unistd.h>
-#else
 #include <stdlib.h>
-#endif /* __APPLE__ */
+#include <unistd.h>
 
 /*
  * The canonical name should be 'go' since we prefer symbol names with fewer
@@ -64,9 +59,5 @@ main(int argc, char **argv)
 {
 	(void) signal(SIGUSR1, handle);
 	for (;;)
-#if !defined(__APPLE__)
-		getpid();
-#else
 		geteuid();
-#endif
 }
