@@ -24,14 +24,8 @@
  * Use is subject to license terms.
  */
 
-#pragma ident	"@(#)tst.weak2.c	1.1	06/08/28 SMI"
-
 #include <signal.h>
-#if !defined(__APPLE__)
-#include <unistd.h>
-#else
 #include <stdlib.h>
-#endif /* __APPLE__ */
 
 /*
  * The canonical name should be 'go' since we prefer symbol names with fewer
