@@ -23,8 +23,6 @@
  * Use is subject to license terms.
  */
 
-#pragma ident	"@(#)ctf.c	1.14	06/05/03 SMI"
-
 /*
  * Create and parse buffers containing CTF data.
  */
@@ -35,7 +33,8 @@
 #include <strings.h>
 #include <ctype.h>
 #include <zlib.h>
-#include <elf.h>
+
+#include <sys/elf.h>
 
 #include "ctf_headers.h"
 #include "ctftools.h"
@@ -48,7 +47,6 @@
 #define SWAP32(v)		v = OSSwapInt32(v)
 #define SWAP64(v)		v = OSSwapInt64(v)
 
-static int needSwap;
 #endif /* __APPLE__ */
 
 /*
@@ -71,6 +69,7 @@ struct ctf_buf {
 	size_t ctb_size;	/* size of buffer */
 	int nptent;		/* number of processed types */
 	int ntholes;		/* number of type holes */
+	int nptrauth;		/* number of ptrauth */
 };
 
 /*PRINTFLIKE1*/
@@ -148,13 +147,6 @@ write_label(labelent_t *le, ctf_buf_t *b)
 
 	ctl.ctl_label = strtab_insert(&b->ctb_strtab, le->le_name);
 	ctl.ctl_typeidx = le->le_idx;
-	
-#if defined(__APPLE__)
-	if (needSwap) {
-		SWAP32(ctl.ctl_label);
-		SWAP32(ctl.ctl_typeidx);
-	}
-#endif /* __APPLE__ */
 
 	ctf_buf_write(b, &ctl, sizeof (ctl));
 
@@ -166,15 +158,9 @@ write_objects(iidesc_t *idp, ctf_buf_t *b)
 {
 	ushort_t id = (idp ? idp->ii_dtype->t_id : 0);
 
-#if defined(__APPLE__)
-	if (needSwap) {
-		SWAP16(id);
-	}
-#endif /* __APPLE__ */
-
 	ctf_buf_write(b, &id, sizeof (id));
 
-	debug(3, "Wrote object %s (%d)\n", (idp ? idp->ii_name : "(null)"), id);
+	debug(3, "Wrote object %s (%d)\n", (idp ? idp->ii_name->value : "(null)"), id);
 }
 
 static void
@@ -197,24 +183,10 @@ write_functions(iidesc_t *idp, ctf_buf_t *b)
 	fdata[0] = CTF_TYPE_INFO(CTF_K_FUNCTION, 1, nargs);
 	fdata[1] = idp->ii_dtype->t_id;
 
-#if defined(__APPLE__)
-	if (needSwap) {
-		SWAP16(fdata[0]);
-		SWAP16(fdata[1]);
-	}
-#endif /* __APPLE__ */
-
 	ctf_buf_write(b, fdata, sizeof (fdata));
 
 	for (i = 0; i < idp->ii_nargs; i++) {
 		id = idp->ii_args[i]->t_id;
-
-#if defined(__APPLE__)
-		if (needSwap) {
-			SWAP16(id);
-		}
-#endif /* __APPLE__ */
-
 		ctf_buf_write(b, &id, sizeof (id));
 	}
 
@@ -223,7 +195,7 @@ write_functions(iidesc_t *idp, ctf_buf_t *b)
 		ctf_buf_write(b, &id, sizeof (id));
 	}
 
-	debug(3, "Wrote function %s (%d args)\n", idp->ii_name, nargs);
+	debug(3, "Wrote function %s (%d args)\n", idp->ii_name->value, nargs);
 }
 
 /*
@@ -239,27 +211,11 @@ write_sized_type_rec(ctf_buf_t *b, ctf_type_t *ctt, size_t size)
 		ctt->ctt_size = CTF_LSIZE_SENT;
 		ctt->ctt_lsizehi = CTF_SIZE_TO_LSIZE_HI(size);
 		ctt->ctt_lsizelo = CTF_SIZE_TO_LSIZE_LO(size);
-#if defined(__APPLE__)
-		if (needSwap) {
-			SWAP32(ctt->ctt_name);
-			SWAP16(ctt->ctt_info);
-			SWAP16(ctt->ctt_size);
-			SWAP32(ctt->ctt_lsizehi);
-			SWAP32(ctt->ctt_lsizelo);
-		}
-#endif /* __APPLE__ */
 		ctf_buf_write(b, ctt, sizeof (*ctt));
 	} else {
 		ctf_stype_t *cts = (ctf_stype_t *)ctt;
 
 		cts->ctt_size = (ushort_t)size;
-#if defined(__APPLE__)
-		if (needSwap) {
-			SWAP32(cts->ctt_name);
-			SWAP16(cts->ctt_info);
-			SWAP16(cts->ctt_size);
-		}
-#endif /* __APPLE__ */
 		ctf_buf_write(b, cts, sizeof (*cts));
 	}
 }
@@ -268,14 +224,6 @@ static void
 write_unsized_type_rec(ctf_buf_t *b, ctf_type_t *ctt)
 {
 	ctf_stype_t *cts = (ctf_stype_t *)ctt;
-
-#if defined(__APPLE__)
-	if (needSwap) {
-		SWAP32(cts->ctt_name);
-		SWAP16(cts->ctt_info);
-		SWAP16(cts->ctt_size);
-	}
-#endif /* __APPLE__ */
 	ctf_buf_write(b, cts, sizeof (*cts));
 }
 
@@ -348,17 +296,13 @@ write_type(tdesc_t *tp, ctf_buf_t *b)
 			encoding = ip->intr_fformat;
 
 		data = CTF_INT_DATA(encoding, ip->intr_offset, ip->intr_nbits);
-#if defined(__APPLE__)
-		if (needSwap) {
-			SWAP32(data);
-		}
-#endif /* __APPLE__ */
 		ctf_buf_write(b, &data, sizeof (data));
 		break;
 
 	case POINTER:
 		ctt.ctt_info = CTF_TYPE_INFO(CTF_K_POINTER, isroot, 0);
 		ctt.ctt_type = tp->t_tdesc->t_id;
+		debug(3, "pointer with %llu\n", tp->t_tdesc->t_id);
 		write_unsized_type_rec(b, &ctt);
 		break;
 
@@ -369,13 +313,6 @@ write_type(tdesc_t *tp, ctf_buf_t *b)
 		cta.cta_contents = tp->t_ardef->ad_contents->t_id;
 		cta.cta_index = tp->t_ardef->ad_idxtype->t_id;
 		cta.cta_nelems = tp->t_ardef->ad_nelems;
-#if defined(__APPLE__)
-		if (needSwap) {
-			SWAP16(cta.cta_contents);
-			SWAP16(cta.cta_index);
-			SWAP32(cta.cta_nelems);
-		}
-#endif /* __APPLE__ */
 		ctf_buf_write(b, &cta, sizeof (cta));
 		break;
 
@@ -400,13 +337,6 @@ write_type(tdesc_t *tp, ctf_buf_t *b)
 				    offset);
 				ctm.ctm_type = mp->ml_type->t_id;
 				ctm.ctm_offset = mp->ml_offset;
-#if defined(__APPLE__)
-				if (needSwap) {
-					SWAP32(ctm.ctm_name);
-					SWAP16(ctm.ctm_type);
-					SWAP16(ctm.ctm_offset);
-				}
-#endif /* __APPLE__ */
 				ctf_buf_write(b, &ctm, sizeof (ctm));
 			}
 		} else {
@@ -421,14 +351,6 @@ write_type(tdesc_t *tp, ctf_buf_t *b)
 				    CTF_OFFSET_TO_LMEMHI(mp->ml_offset);
 				ctlm.ctlm_offsetlo =
 				    CTF_OFFSET_TO_LMEMLO(mp->ml_offset);
-#if defined(__APPLE__)
-				if (needSwap) {
-					SWAP32(ctlm.ctlm_name);
-					SWAP16(ctlm.ctlm_type);
-					SWAP32(ctlm.ctlm_offsethi);
-					SWAP32(ctlm.ctlm_offsetlo);
-				}
-#endif /* __APPLE__ */
 				ctf_buf_write(b, &ctlm, sizeof (ctlm));
 			}
 		}
@@ -445,12 +367,6 @@ write_type(tdesc_t *tp, ctf_buf_t *b)
 			offset = strtab_insert(&b->ctb_strtab, ep->el_name);
 			cte.cte_name = CTF_TYPE_NAME(CTF_STRTAB_0, offset);
 			cte.cte_value = ep->el_number;
-#if defined(__APPLE__)
-			if (needSwap) {
-				SWAP32(cte.cte_name);
-				SWAP32(cte.cte_value);
-			}
-#endif /* __APPLE__ */
 			ctf_buf_write(b, &cte, sizeof (cte));
 		}
 		break;
@@ -487,11 +403,6 @@ write_type(tdesc_t *tp, ctf_buf_t *b)
 
 		for (i = 0; i < tp->t_fndef->fn_nargs; i++) {
 			id = tp->t_fndef->fn_args[i]->t_id;
-#if defined(__APPLE__)
-			if (needSwap) {
-				SWAP16(id);
-			}
-#endif /* __APPLE__ */
 			ctf_buf_write(b, &id, sizeof (id));
 		}
 
@@ -513,6 +424,17 @@ write_type(tdesc_t *tp, ctf_buf_t *b)
 		write_unsized_type_rec(b, &ctt);
 		break;
 
+	case PTRAUTH:
+		ctt.ctt_info = CTF_TYPE_INFO(CTF_K_PTRAUTH, isroot, 1);
+		ctt.ctt_type = tp->t_ptrauth->pta_type->t_id;
+		write_unsized_type_rec(b, &ctt);
+
+		data = CTF_PTRAUTH_DATA(tp->t_ptrauth->pta_key, tp->t_ptrauth->pta_discriminator, tp->t_ptrauth->pta_discriminated);
+		ctf_buf_write(b, &data, sizeof (data));
+
+		b->nptrauth++;
+
+		break;
 	default:
 		warning("Can't write unknown type %d\n", tp->t_type);
 	}
@@ -684,11 +606,6 @@ ctf_gen(iiburst_t *iiburst, size_t *resszp, int do_compress)
 	caddr_t outbuf;
 
 	int i;
-	
-#if defined(__APPLE__)
-	needSwap = (do_compress & CTF_BYTESWAP);
-	do_compress &= ~CTF_BYTESWAP;
-#endif /* __APPLE__ */
 
 	/*
 	 * Prepare the header, and create the CTF output buffers.  The data
@@ -696,7 +613,6 @@ ctf_gen(iiburst_t *iiburst, size_t *resszp, int do_compress)
 	 * integers; we pad these out to the next 4-byte boundary if needed.
 	 */
 	h.cth_magic = CTF_MAGIC;
-	h.cth_version = CTF_VERSION;
 	h.cth_flags = do_compress ? CTF_F_COMPRESS : 0;
 	h.cth_parlabel = strtab_insert(&buf->ctb_strtab,
 	    iiburst->iib_td->td_parlabel);
@@ -704,8 +620,7 @@ ctf_gen(iiburst_t *iiburst, size_t *resszp, int do_compress)
 	    iiburst->iib_td->td_parname);
 
 	h.cth_lbloff = 0;
-	(void) list_iter(iiburst->iib_td->td_labels, (int (*)())write_label,
-	    buf);
+	tdata_label_iter(iiburst->iib_td, (int (*)())write_label, buf);
 
 	pad_buffer(buf, 2);
 	h.cth_objtoff = ctf_buf_cur(buf);
@@ -719,26 +634,18 @@ ctf_gen(iiburst_t *iiburst, size_t *resszp, int do_compress)
 
 	pad_buffer(buf, 4);
 	h.cth_typeoff = ctf_buf_cur(buf);
-	(void) list_iter(iiburst->iib_types, (int (*)())write_type, buf);
+	array_sort(iiburst->iib_types, tdesc_idcmp);
+	(void) array_iter(iiburst->iib_types, (int (*)())write_type, buf);
 
-	debug(2, "CTF wrote %d types\n", list_count(iiburst->iib_types));
+	debug(2, "CTF wrote %d types\n", array_count(iiburst->iib_types));
 
 	h.cth_stroff = ctf_buf_cur(buf);
 	h.cth_strlen = strtab_size(&buf->ctb_strtab);
-	
-#if defined(__APPLE__)
-	if (needSwap) {
-		SWAP16(h.cth_preamble.ctp_magic);
-		SWAP32(h.cth_parlabel);	/* ref to name of parent lbl uniq'd against */
-		SWAP32(h.cth_parname);	/* ref to basename of parent */
-		SWAP32(h.cth_lbloff);	/* offset of label section */
-		SWAP32(h.cth_objtoff);	/* offset of object section */
-		SWAP32(h.cth_funcoff);	/* offset of function section */
-		SWAP32(h.cth_typeoff);	/* offset of type section */
-		SWAP32(h.cth_stroff);	/* offset of string section */
-		SWAP32(h.cth_strlen);	/* length of string section in bytes */
-	}
-#endif /* __APPLE__ */
+
+	/*
+	 * Encoded pointers are only supported in V3 or later
+	 */
+	h.cth_version = buf->nptrauth > 0 ? CTF_VERSION : CTF_VERSION_2;
 
 	/*
 	 * We only do compression for ctfmerge, as ctfconvert is only
@@ -786,6 +693,9 @@ count_types(ctf_header_t *h, caddr_t data)
 		case CTF_K_FLOAT:
 			dptr += 4;
 			break;
+		case CTF_K_PTRAUTH:
+			dptr += 4;
+			break;
 		case CTF_K_POINTER:
 		case CTF_K_FORWARD:
 		case CTF_K_TYPEDEF:
@@ -910,18 +820,18 @@ resurrect_objects(ctf_header_t *h, tdata_t *td, tdesc_t **tdarr, int tdsize,
 			parseterminate("Reference to invalid type %d", id);
 		}
 
-		ii = iidesc_new(symit_name(si));
+		ii = iidesc_new(atom_get(symit_name(si)));
 		ii->ii_dtype = tdarr[id];
 		if (GELF_ST_BIND(sym->st_info) == STB_LOCAL) {
 			ii->ii_type = II_SVAR;
-			ii->ii_owner = xstrdup(symit_curfile(si));
+			ii->ii_owner = atom_get(symit_curfile(si));
 		} else
 			ii->ii_type = II_GVAR;
 		hash_add(td->td_iihash, ii);
 
 		debug(3, "Resurrected %s object %s (%d) from %s\n",
 		    (ii->ii_type == II_GVAR ? "global" : "static"),
-		    ii->ii_name, id, (ii->ii_owner ? ii->ii_owner : "(none)"));
+		    ii->ii_name->value, id, atom_pretty(ii->ii_owner, "(none)"));
 	}
 }
 
@@ -960,11 +870,11 @@ resurrect_functions(ctf_header_t *h, tdata_t *td, tdesc_t **tdarr, int tdsize,
 		if (retid >= tdsize)
 			parseterminate("Reference to invalid type %d", retid);
 
-		ii = iidesc_new(symit_name(si));
+		ii = iidesc_new(atom_get(symit_name(si)));
 		ii->ii_dtype = tdarr[retid];
 		if (GELF_ST_BIND(sym->st_info) == STB_LOCAL) {
 			ii->ii_type = II_SFUN;
-			ii->ii_owner = xstrdup(symit_curfile(si));
+			ii->ii_owner = atom_get(symit_curfile(si));
 		} else
 			ii->ii_type = II_GFUN;
 		ii->ii_nargs = CTF_INFO_VLEN(info);
@@ -990,7 +900,7 @@ resurrect_functions(ctf_header_t *h, tdata_t *td, tdesc_t **tdarr, int tdsize,
 
 		debug(3, "Resurrected %s function %s (%d, %d args)\n",
 		    (ii->ii_type == II_GFUN ? "global" : "static"),
-		    ii->ii_name, retid, ii->ii_nargs);
+		    ii->ii_name->value, retid, ii->ii_nargs);
 	}
 }
 
@@ -1010,11 +920,13 @@ resurrect_types(ctf_header_t *h, tdata_t *td, tdesc_t **tdarr, int tdsize,
 	int iicnt = 0;
 	tid_t tid, argid;
 	int kind, vlen;
+	int nargs, vargs;
 	int i;
 
 	elist_t **epp;
 	mlist_t **mpp;
 	intr_t *ip;
+	ptrauth_t *pta;
 
 	ctf_type_t *ctt;
 	ctf_array_t *cta;
@@ -1047,9 +959,9 @@ resurrect_types(ctf_header_t *h, tdata_t *td, tdesc_t **tdarr, int tdsize,
 				"Unable to cope with non-zero strtab id");
 		if (CTF_NAME_OFFSET(ctt->ctt_name) != 0) {
 			tdp->t_name =
-			    xstrdup(sbuf + CTF_NAME_OFFSET(ctt->ctt_name));
+			    atom_get(sbuf + CTF_NAME_OFFSET(ctt->ctt_name));
 		} else
-			tdp->t_name = NULL;
+			tdp->t_name = ATOM_NULL;
 
 		kind = CTF_INFO_KIND(ctt->ctt_info);
 		vlen = CTF_INFO_VLEN(ctt->ctt_info);
@@ -1131,7 +1043,7 @@ resurrect_types(ctf_header_t *h, tdata_t *td, tdesc_t **tdarr, int tdsize,
 					dptr += sizeof (ctf_member_t);
 
 					*mpp = xmalloc(sizeof (mlist_t));
-					(*mpp)->ml_name = xstrdup(sbuf +
+					(*mpp)->ml_name = atom_get(sbuf +
 					    ctm->ctm_name);
 					(*mpp)->ml_type = tdarr[ctm->ctm_type];
 					(*mpp)->ml_offset = ctm->ctm_offset;
@@ -1146,7 +1058,7 @@ resurrect_types(ctf_header_t *h, tdata_t *td, tdesc_t **tdarr, int tdsize,
 					dptr += sizeof (ctf_lmember_t);
 
 					*mpp = xmalloc(sizeof (mlist_t));
-					(*mpp)->ml_name = xstrdup(sbuf +
+					(*mpp)->ml_name = atom_get(sbuf +
 					    ctlm->ctlm_name);
 					(*mpp)->ml_type =
 					    tdarr[ctlm->ctlm_type];
@@ -1170,7 +1082,7 @@ resurrect_types(ctf_header_t *h, tdata_t *td, tdesc_t **tdarr, int tdsize,
 				dptr += sizeof (ctf_enum_t);
 
 				*epp = xmalloc(sizeof (elist_t));
-				(*epp)->el_name = xstrdup(sbuf + cte->cte_name);
+				(*epp)->el_name = atom_get(sbuf + cte->cte_name);
 				(*epp)->el_number = cte->cte_value;
 			}
 			*epp = NULL;
@@ -1178,7 +1090,6 @@ resurrect_types(ctf_header_t *h, tdata_t *td, tdesc_t **tdarr, int tdsize,
 
 		case CTF_K_FORWARD:
 			tdp->t_type = FORWARD;
-			list_add(&td->td_fwdlist, tdp);
 			break;
 
 		case CTF_K_TYPEDEF:
@@ -1198,17 +1109,18 @@ resurrect_types(ctf_header_t *h, tdata_t *td, tdesc_t **tdarr, int tdsize,
 
 		case CTF_K_FUNCTION:
 			tdp->t_type = FUNCTION;
-			tdp->t_fndef = xcalloc(sizeof (fndef_t));
-			tdp->t_fndef->fn_ret = tdarr[ctt->ctt_type];
 
 			/* LINTED - pointer alignment */
+			vargs = nargs = 0;
 			if (vlen > 0 && *(ushort_t *)(dptr +
 			    (sizeof (ushort_t) * (vlen - 1))) == 0)
-				tdp->t_fndef->fn_vargs = 1;
+				vargs = 1;
+			nargs = vlen - vargs;
 
-			tdp->t_fndef->fn_nargs = vlen - tdp->t_fndef->fn_vargs;
-			tdp->t_fndef->fn_args = xcalloc(sizeof (tdesc_t) *
-			    vlen - tdp->t_fndef->fn_vargs);
+			tdp->t_fndef = xcalloc(sizeof (fndef_t) + sizeof(tdesc_t *) * nargs);
+			tdp->t_fndef->fn_ret = tdarr[ctt->ctt_type];
+			tdp->t_fndef->fn_vargs = vargs;
+			tdp->t_fndef->fn_nargs = nargs;
 
 			for (i = 0; i < vlen; i++) {
 				/* LINTED - pointer alignment */
@@ -1228,6 +1140,22 @@ resurrect_types(ctf_header_t *h, tdata_t *td, tdesc_t **tdarr, int tdsize,
 			tdp->t_tdesc = tdarr[ctt->ctt_type];
 			break;
 
+		case CTF_K_PTRAUTH:
+			tdp->t_type = PTRAUTH;
+
+			/* LINTED - pointer alignment */
+			data = *((uint_t *)dptr);
+			dptr += sizeof (uint_t);
+
+			pta = xcalloc(sizeof (ptrauth_t));
+			pta->pta_discriminator = CTF_PTRAUTH_DISCRIMINATOR(data);
+			pta->pta_discriminated = CTF_PTRAUTH_DISCRIMINATED(data);
+			pta->pta_key = CTF_PTRAUTH_KEY(data);
+			pta->pta_type = tdarr[ctt->ctt_type];
+
+			tdp->t_ptrauth = pta;
+			break;
+
 		case CTF_K_UNKNOWN:
 			break;
 
@@ -1278,7 +1206,7 @@ ctf_parse(ctf_header_t *h, caddr_t buf, symit_data_t *si, char *label)
 		tdarr[i]->t_id = i;
 	}
 
-	td->td_parlabel = xstrdup(buf + h->cth_stroff + h->cth_parlabel);
+	td->td_parlabel = atom_get(buf + h->cth_stroff + h->cth_parlabel);
 
 	/* we have the technology - we can rebuild them */
 	idx = resurrect_labels(h, td, buf, label);
@@ -1350,7 +1278,7 @@ ctf_load(char *file, caddr_t buf, size_t bufsz, symit_data_t *si, char *label)
 	if (h->cth_magic != CTF_MAGIC)
 		parseterminate("Corrupt CTF - bad magic 0x%x", h->cth_magic);
 
-	if (h->cth_version != CTF_VERSION)
+	if (h->cth_version != CTF_VERSION_2 && h->cth_version != CTF_VERSION_3)
 		parseterminate("Unknown CTF version %d", h->cth_version);
 
 	ctfdatasz = h->cth_stroff + h->cth_strlen;
