@@ -23,8 +23,6 @@
  * Use is subject to license terms.
  */
 
-#pragma ident	"@(#)dwarf.c	1.9	07/01/10 SMI"
-
 /*
  * DWARF to tdata conversion
  *
@@ -82,6 +80,7 @@
  * traversals.
  */
 
+#include <assert.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <strings.h>
@@ -95,16 +94,22 @@
 #include "ctftools.h"
 #include "memory.h"
 #include "list.h"
+#include "alist.h"
 #include "traverse.h"
 
 #if defined(__APPLE__)
 /* Sun extensions not present in Darwin's dwarf.h */
 #define DW_ATE_SUN_interval_float       0x91
 #define DW_ATE_SUN_imaginary_float      0x92 /* Obsolete: See DW_ATE_imaginary_float */
+
+/* it is EXTREMELY slow, and doesn't affect memory that much */
+#define dwarf_dealloc(...)
 #endif /* __APPLE__ */
 
 /* The version of DWARF which we support. */
 #define	DWARF_VERSION	2
+#define	DWARF_VERSION3	3
+#define	DWARF_VERSION4	4
 
 /*
  * We need to define a couple of our own intrinsics, to smooth out some of the
@@ -135,8 +140,9 @@ typedef struct dwarf {
 	Dwarf_Debug dw_dw;		/* for libdwarf */
 	Dwarf_Error dw_err;		/* for libdwarf */
 	Dwarf_Unsigned dw_maxoff;	/* highest legal offset in this cu */
+	Dwarf_Off dw_cuoff;		/* lowest offset in this cu */
 	tdata_t *dw_td;			/* root of the tdesc/iidesc tree */
-	hash_t *dw_tidhash;		/* hash of tdescs by t_id */
+	alist_t *dw_tidhash;		/* hash of tdescs by t_id */
 	hash_t *dw_fwdhash;		/* hash of fwd decls by name */
 	hash_t *dw_enumhash;		/* hash of memberless enums by name */
 	tdesc_t *dw_void;		/* manufactured void type */
@@ -144,10 +150,19 @@ typedef struct dwarf {
 	size_t dw_ptrsz;		/* size of a pointer in this file */
 	tid_t dw_mfgtid_last;		/* last mfg'd type ID used */
 	uint_t dw_nunres;		/* count of unresolved types */
-	char *dw_cuname;		/* name of compilation unit */
+	Dwarf_Die dw_cu;		/* die of cu */
+	atom_t *dw_cuname;		/* name of compilation unit */
 } dwarf_t;
 
-static void die_create_one(dwarf_t *, Dwarf_Die);
+typedef struct cu_data
+{
+	struct cu_data *cud_next;
+	atom_t *cud_name;
+	Dwarf_Unsigned cud_nxthdr;
+} cu_data_t;
+
+
+static tdesc_t * die_create_one(dwarf_t *, Dwarf_Die);
 static void die_create(dwarf_t *, Dwarf_Die);
 
 static tid_t
@@ -159,25 +174,20 @@ mfgtid_next(dwarf_t *dw)
 static void
 tdesc_add(dwarf_t *dw, tdesc_t *tdp)
 {
-	hash_add(dw->dw_tidhash, tdp);
+	alist_add(dw->dw_tidhash, (void *)(uintptr_t)tdp->t_id, tdp);
 }
 
 static tdesc_t *
 tdesc_lookup(dwarf_t *dw, int tid)
 {
-	tdesc_t tmpl, *tdp;
-
-	tmpl.t_id = tid;
-
-	if (hash_find(dw->dw_tidhash, &tmpl, (void **)&tdp))
-		return (tdp);
-	else
-		return (NULL);
+	void *tpd = NULL;
+	(void)alist_find(dw->dw_tidhash, (void *)(uintptr_t)tid, &tpd);
+	return tpd;
 }
 
 /*
  * Resolve a tdesc down to a node which should have a size.  Returns the size,
- * zero if the size hasn't yet been determined.
+ * SIZE_MAX if the size hasn't yet been determined.
  */
 static size_t
 tdesc_size(tdesc_t *tdp)
@@ -191,10 +201,11 @@ tdesc_size(tdesc_t *tdp)
 		case STRUCT:
 		case UNION:
 		case ENUM:
+		case PTRAUTH:
 			return (tdp->t_size);
 
 		case FORWARD:
-			return (0);
+			return (SIZE_MAX);
 
 		case TYPEDEF:
 		case VOLATILE:
@@ -204,7 +215,7 @@ tdesc_size(tdesc_t *tdp)
 			continue;
 
 		case 0: /* not yet defined */
-			return (0);
+			return (SIZE_MAX);
 
 		default:
 			terminate("tdp %u: tdesc_size on unknown type %d\n",
@@ -227,6 +238,7 @@ tdesc_bitsize(tdesc_t *tdp)
 		case UNION:
 		case ENUM:
 		case POINTER:
+		case PTRAUTH:
 			return (tdp->t_size * NBBY);
 
 		case FORWARD:
@@ -378,6 +390,17 @@ die_signed(dwarf_t *dw, Dwarf_Die die, Dwarf_Half name, Dwarf_Signed *valp,
 	if ((attr = die_attr(dw, die, name, req)) == NULL)
 		return (0); /* die_attr will terminate for us if necessary */
 
+	if (!dwarf_formisdata(attr)) {
+		if (req) {
+			terminate("die %llu: is not data (form 0x%x)\n",
+			    die_off(dw, die), die_attr_form(dw, attr));
+		}
+		else {
+			dwarf_dealloc(dw->dw_dw, attr, DW_DLA_ATTR);
+			return (0);
+		}
+	}
+
 	if (dwarf_formsdata(attr, &val, &dw->dw_err) != DW_DLV_OK) {
 		terminate("die %llu: failed to get signed (form 0x%x)\n",
 		    die_off(dw, die), die_attr_form(dw, attr));
@@ -399,6 +422,17 @@ die_unsigned(dwarf_t *dw, Dwarf_Die die, Dwarf_Half name, Dwarf_Unsigned *valp,
 	if ((attr = die_attr(dw, die, name, req)) == NULL)
 		return (0); /* die_attr will terminate for us if necessary */
 
+	if (!dwarf_formisdata(attr)) {
+		if (req) {
+			terminate("die %llu: is not data (form 0x%x)\n",
+			    die_off(dw, die), die_attr_form(dw, attr));
+		}
+		else {
+			dwarf_dealloc(dw->dw_dw, attr, DW_DLA_ATTR);
+			return (0);
+		}
+	}
+
 	if (dwarf_formudata(attr, &val, &dw->dw_err) != DW_DLV_OK) {
 		terminate("die %llu: failed to get unsigned (form 0x%x)\n",
 		    die_off(dw, die), die_attr_form(dw, attr));
@@ -421,7 +455,7 @@ die_bool(dwarf_t *dw, Dwarf_Die die, Dwarf_Half name, Dwarf_Bool *valp, int req)
 
 	if (dwarf_formflag(attr, &val, &dw->dw_err) != DW_DLV_OK) {
 		terminate("die %llu: failed to get bool (form 0x%x)\n",
-		    die_off(dw, die), die_attr_form(dw, attr));
+			die_off(dw, die), die_attr_form(dw, attr));
 	}
 
 	dwarf_dealloc(dw->dw_dw, attr, DW_DLA_ATTR);
@@ -431,25 +465,38 @@ die_bool(dwarf_t *dw, Dwarf_Die die, Dwarf_Half name, Dwarf_Bool *valp, int req)
 }
 
 static int
-die_string(dwarf_t *dw, Dwarf_Die die, Dwarf_Half name, char **strp, int req)
+die_atom(dwarf_t *dw, Dwarf_Die die, Dwarf_Half name, atom_t **atomp, int req)
 {
 	Dwarf_Attribute attr;
 	char *str;
 
-	if ((attr = die_attr(dw, die, name, req)) == NULL)
+	if ((attr = die_attr(dw, die, name, req)) == NULL) {
+		*atomp = ATOM_NULL;
 		return (0); /* die_attr will terminate for us if necessary */
+	}
 
 	if (dwarf_formstring(attr, &str, &dw->dw_err) != DW_DLV_OK) {
 		terminate("die %llu: failed to get string (form 0x%x)\n",
-		    die_off(dw, die), die_attr_form(dw, attr));
+				  die_off(dw, die), die_attr_form(dw, attr));
 	}
 
-	*strp = xstrdup(str);
+	*atomp = atom_get(str);
 	dwarf_dealloc(dw->dw_dw, str, DW_DLA_STRING);
 
 	return (1);
 }
 
+static int
+die_string(dwarf_t *dw, Dwarf_Die die, Dwarf_Half name, char **strp, int req)
+{
+	atom_t *atom;
+	int rc;
+
+	rc = die_atom(dw, die, name, &atom, req);
+	*strp = atom ? xstrdup(atom->value) : NULL;
+	return rc;
+}
+
 static Dwarf_Off
 die_attr_ref(dwarf_t *dw, Dwarf_Die die, Dwarf_Half name)
 {
@@ -458,9 +505,9 @@ die_attr_ref(dwarf_t *dw, Dwarf_Die die, Dwarf_Half name)
 
 	attr = die_attr(dw, die, name, DW_ATTR_REQ);
 
-	if (dwarf_formref(attr, &off, &dw->dw_err) != DW_DLV_OK) {
+	if (dwarf_global_formref(attr, &off, &dw->dw_err) != DW_DLV_OK) {
 		terminate("die %llu: failed to get ref (form 0x%x)\n",
-		    die_off(dw, die), die_attr_form(dw, attr));
+			die_off(dw, die), die_attr_form(dw, attr));
 	}
 
 	dwarf_dealloc(dw->dw_dw, attr, DW_DLA_ATTR);
@@ -468,14 +515,14 @@ die_attr_ref(dwarf_t *dw, Dwarf_Die die, Dwarf_Half name)
 	return (off);
 }
 
-static char *
+static atom_t *
 die_name(dwarf_t *dw, Dwarf_Die die)
 {
-	char *str = NULL;
+	atom_t *atom = ATOM_NULL;
 
-	(void) die_string(dw, die, DW_AT_name, &str, 0);
+	(void) die_atom(dw, die, DW_AT_name, &atom, 0);
 
-	return (str);
+	return (atom);
 }
 
 static int
@@ -487,14 +534,14 @@ die_isdecl(dwarf_t *dw, Dwarf_Die die)
 }
 
 #if defined(__APPLE__)
-static char *
+static atom_t *
 die_linkage_name(dwarf_t *dw, Dwarf_Die die)
 {
-	char *str = NULL;
+	atom_t *atom = ATOM_NULL;
 
-	(void) die_string(dw, die, DW_AT_MIPS_linkage_name, &str, 0);
+	(void) die_atom(dw, die, DW_AT_MIPS_linkage_name, &atom, 0);
 
-	return (str);
+	return (atom);
 }
 
 static Dwarf_Die
@@ -563,6 +610,17 @@ die_lookup_pass1(dwarf_t *dw, Dwarf_Die die, Dwarf_Half name)
 	if ((tdp = tdesc_lookup(dw, ref)) != NULL)
 		return (tdp);
 
+	if ((ref < dw->dw_cuoff) || (ref > dw->dw_maxoff))
+	{
+	    Dwarf_Die ref_die;
+		int rc;
+		if ((rc = dwarf_offdie(dw->dw_dw, ref, &ref_die, &dw->dw_err)) != DW_DLV_OK) {
+			terminate("die %llu: failed to find oob ref\n", die);
+		}
+		tdp = die_create_one(dw, ref_die);
+	    return (tdp);
+	}
+
 	return (die_add(dw, ref));
 }
 
@@ -573,26 +631,50 @@ die_mem_offset(dwarf_t *dw, Dwarf_Die die, Dwarf_Half name,
 	Dwarf_Attribute attr;
 	Dwarf_Locdesc *loc;
 	Dwarf_Signed locnum;
+	Dwarf_Half form;
 
 	if ((attr = die_attr(dw, die, name, req)) == NULL)
 		return (0); /* die_attr will terminate for us if necessary */
 
-	if (dwarf_loclist(attr, &loc, &locnum, &dw->dw_err) != DW_DLV_OK) {
-		terminate("die %llu: failed to get mem offset location list\n",
-		    die_off(dw, die));
+	if (dwarf_whatform(attr, &form, &dw->dw_err) != DW_DLV_OK)
+		return (0);
+
+	switch (form) {
+	case DW_FORM_sec_offset:
+	case DW_FORM_block:
+	case DW_FORM_block1:
+	case DW_FORM_block2:
+	case DW_FORM_block4:
+		if (dwarf_loclist(attr, &loc, &locnum, &dw->dw_err) != DW_DLV_OK) {
+			terminate("die %llu: failed to get mem offset location list\n",
+			    die_off(dw, die));
+		}
+
+		dwarf_dealloc(dw->dw_dw, attr, DW_DLA_ATTR);
+
+		if (locnum != 1 || loc->ld_s->lr_atom != DW_OP_plus_uconst) {
+			terminate("die %llu: cannot parse member offset\n",
+			    die_off(dw, die));
+		}
+
+		*valp = loc->ld_s->lr_number;
+
+		dwarf_dealloc(dw->dw_dw, loc->ld_s, DW_DLA_LOC_BLOCK);
+		dwarf_dealloc(dw->dw_dw, loc, DW_DLA_LOCDESC);
+		break;
+	case DW_FORM_data1:
+	case DW_FORM_data2:
+	case DW_FORM_data4:
+	case DW_FORM_data8:
+	case DW_FORM_udata:
+		if (!die_unsigned(dw, die, name, valp, req))
+			return 0;
+		break;
+	default:
+		terminate("die %llu: cannot parse member offset with form "
+			"%u\n", die_off(dw, die), form);
 	}
 
-	dwarf_dealloc(dw->dw_dw, attr, DW_DLA_ATTR);
-
-	if (locnum != 1 || loc->ld_s->lr_atom != DW_OP_plus_uconst) {
-		terminate("die %llu: cannot parse member offset\n",
-		    die_off(dw, die));
-	}
-
-	*valp = loc->ld_s->lr_number;
-
-	dwarf_dealloc(dw->dw_dw, loc->ld_s, DW_DLA_LOC_BLOCK);
-	dwarf_dealloc(dw->dw_dw, loc, DW_DLA_LOCDESC);
 
 	return (1);
 }
@@ -609,7 +691,7 @@ tdesc_intr_common(dwarf_t *dw, int tid, const char *name, size_t sz)
 	intr->intr_nbits = sz * NBBY;
 
 	tdp = xcalloc(sizeof (tdesc_t));
-	tdp->t_name = xstrdup(name);
+	tdp->t_name = atom_get(name);
 	tdp->t_size = sz;
 	tdp->t_id = tid;
 	tdp->t_type = INTRINSIC;
@@ -665,7 +747,7 @@ tdesc_intr_clone(dwarf_t *dw, tdesc_t *old, size_t bitsz)
 		    "unresolved type\n", old->t_id);
 	}
 
-	new->t_name = xstrdup(old->t_name);
+	new->t_name = old->t_name;
 	new->t_size = old->t_size;
 	new->t_id = mfgtid_next(dw);
 	new->t_type = INTRINSIC;
@@ -721,6 +803,10 @@ tdesc_array_create(dwarf_t *dw, Dwarf_Die dim, tdesc_t *arrtdp,
 		ar->ad_nelems = uval + 1;
 	else if (die_signed(dw, dim, DW_AT_upper_bound, &sval, 0))
 		ar->ad_nelems = sval + 1;
+	else if (die_unsigned(dw, dim, DW_AT_count, &uval, 0))
+		ar->ad_nelems = uval;
+	else if (die_signed(dw, dim, DW_AT_count, &sval, 0))
+		ar->ad_nelems = sval;
 	else
 		ar->ad_nelems = 0;
 
@@ -797,7 +883,7 @@ die_array_resolve(tdesc_t *tdp, tdesc_t **tdpp, void *private)
 	debug(3, "trying to resolve array %d (cont %d)\n", tdp->t_id,
 	    tdp->t_ardef->ad_contents->t_id);
 
-	if ((sz = tdesc_size(tdp->t_ardef->ad_contents)) == 0) {
+	if ((sz = tdesc_size(tdp->t_ardef->ad_contents)) == SIZE_MAX) {
 		debug(3, "unable to resolve array %s (%d) contents %d\n",
 		    tdesc_name(tdp), tdp->t_id,
 		    tdp->t_ardef->ad_contents->t_id);
@@ -854,6 +940,7 @@ die_enum_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 
 		do {
 			elist_t *el;
+			Dwarf_Attribute attr;
 
 			if (die_tag(dw, mem) != DW_TAG_enumerator) {
 				/* Nested type declaration */
@@ -864,7 +951,11 @@ die_enum_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 			el = xcalloc(sizeof (elist_t));
 			el->el_name = die_name(dw, mem);
 
-			if (die_signed(dw, mem, DW_AT_const_value, &sval, 0)) {
+			int is_unsigned = 0;
+			if ((attr = die_attr(dw, mem, DW_AT_const_value, 0)) != NULL)
+				is_unsigned = dwarf_formisudata(attr);
+
+			if (!is_unsigned && die_signed(dw, mem, DW_AT_const_value, &sval, 0)) {
 				el->el_number = sval;
 			} else if (die_unsigned(dw, mem, DW_AT_const_value,
 			    &uval, 0)) {
@@ -875,7 +966,7 @@ die_enum_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 			}
 
 			debug(3, "die %llu: enum %llu: created %s = %d\n", off,
-			    die_off(dw, mem), el->el_name, el->el_number);
+			    die_off(dw, mem), el->el_name->value, el->el_number);
 
 			*elastp = el;
 			elastp = &el->el_next;
@@ -886,10 +977,10 @@ die_enum_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 
 		tdp->t_flags |= TDESC_F_RESOLVED;
 
-		if (tdp->t_name != NULL) {
+		if (tdp->t_name != ATOM_NULL) {
 			iidesc_t *ii = xcalloc(sizeof (iidesc_t));
 			ii->ii_type = II_SOU;
-			ii->ii_name = xstrdup(tdp->t_name);
+			ii->ii_name = tdp->t_name;
 			ii->ii_dtype = tdp;
 
 			iidesc_add(dw->dw_td->td_iihash, ii);
@@ -1012,14 +1103,15 @@ die_sou_create(dwarf_t *dw, Dwarf_Die str, Dwarf_Off off, tdesc_t *tdp,
 		 * bug 11816).
 		 */
 		if ((ml->ml_name = die_name(dw, mem)) == NULL)
-			ml->ml_name = "";
+			ml->ml_name = atom_get("");
 
+        Dwarf_Off ref = die_attr_ref(dw, mem, DW_AT_type);
 		ml->ml_type = die_lookup_pass1(dw, mem, DW_AT_type);
 
 		if (die_mem_offset(dw, mem, DW_AT_data_member_location,
 		    &mloff, 0)) {
-			debug(3, "die %llu: got mloff %llx\n", off,
-			    (u_longlong_t)mloff);
+			debug(3, "die %llu: got mloff %llx ref %lld\n", off,
+			    (u_longlong_t)mloff, ref);
 			ml->ml_offset = mloff * 8;
 		}
 
@@ -1032,13 +1124,41 @@ die_sou_create(dwarf_t *dw, Dwarf_Die str, Dwarf_Off off, tdesc_t *tdp,
 #ifdef	_BIG_ENDIAN
 			ml->ml_offset += bitoff;
 #else
-			ml->ml_offset += tdesc_bitsize(ml->ml_type) - bitoff -
-			    ml->ml_size;
+			/*
+			 * A DW_AT_byte_size attribute whose value (see Section 2.19) is the
+			 * number of bytes that contain an instance of the bit field and any
+			 * padding bits.
+			 *
+			 * The byte size attribute may be omitted if the size of the object
+			 * containing the bit field can be inferred from the type attribute of
+			 * the data member containing the bit field.
+			 *
+			 * There is some cases where we don't know, yet, the
+			 * size of some intrinsic types.  We must be sure that
+			 * in this case we have a AT_byte_size attribute.
+			 */
+			Dwarf_Unsigned bytesz = 0;
+			if (die_unsigned(dw, mem, DW_AT_byte_size, &bytesz, 0)) {
+				ml->ml_offset += (bytesz * 8) - bitoff - ml->ml_size;
+			} else {
+				size_t tdbitsz = tdesc_bitsize(ml->ml_type);
+				assert((tdbitsz != 0) && "AT_byte_size missing and cannot figure the bitfield size");
+
+				ml->ml_offset += tdbitsz - bitoff - ml->ml_size;
+			}
 #endif
 		}
+		/*
+		 * Clang generates a DW_AT_data_bit_offset attribute for bitfields
+		 * that stores the number of bits from the beginning of the containing
+		 * struct to the beginning of the data member
+		 */
+		else if (die_unsigned(dw, mem, DW_AT_data_bit_offset, &bitoff, 0)) {
 
+			ml->ml_offset += bitoff;
+		}
 		debug(3, "die %llu: mem %llu: created \"%s\" (off %u sz %u)\n",
-		    off, memoff, ml->ml_name, ml->ml_offset, ml->ml_size);
+		    off, memoff, ml->ml_name->value, ml->ml_offset, ml->ml_size);
 
 		*mlastp = ml;
 		mlastp = &ml->ml_next;
@@ -1072,17 +1192,15 @@ die_sou_create(dwarf_t *dw, Dwarf_Die str, Dwarf_Off off, tdesc_t *tdp,
 
 		debug(3, "die %llu: worked around %s %s\n", off, typename, old);
 
-		if (tdp->t_name != NULL)
-			free(tdp->t_name);
-		tdp->t_name = new;
+		tdp->t_name = atom_get_consume(new);
 		return;
 	}
 
 out:
-	if (tdp->t_name != NULL) {
+	if (tdp->t_name != ATOM_NULL) {
 		ii = xcalloc(sizeof (iidesc_t));
 		ii->ii_type = II_SOU;
-		ii->ii_name = xstrdup(tdp->t_name);
+		ii->ii_name = tdp->t_name;
 		ii->ii_dtype = tdp;
 
 		iidesc_add(dw->dw_td->td_iihash, ii);
@@ -1187,7 +1305,7 @@ die_sou_failed(tdesc_t *tdp, tdesc_t **tdpp, void *private)
 		if (ml->ml_size == 0) {
 			fprintf(stderr, "%s %d: failed to size member \"%s\" "
 			    "of type %s (%d)\n", typename, tdp->t_id,
-			    ml->ml_name, tdesc_name(ml->ml_type),
+			    ml->ml_name->value, tdesc_name(ml->ml_type),
 			    ml->ml_type->t_id);
 		}
 	}
@@ -1201,7 +1319,8 @@ die_funcptr_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 	Dwarf_Attribute attr;
 	Dwarf_Half tag;
 	Dwarf_Die arg;
-	fndef_t *fn;
+	fndef_t fnbuf = {0};
+	fndef_t *fn = &fnbuf;
 	int i;
 
 	debug(3, "die %llu: creating function pointer\n", off);
@@ -1229,8 +1348,6 @@ die_funcptr_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 		return;
 	}
 
-	fn = xcalloc(sizeof (fndef_t));
-
 	tdp->t_type = FUNCTION;
 
 	if ((attr = die_attr(dw, die, DW_AT_type, 0)) != NULL) {
@@ -1252,11 +1369,13 @@ die_funcptr_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 			fn->fn_vargs = 1;
 	}
 
+	fn = xcalloc(sizeof(fndef_t) + fn->fn_nargs * sizeof(tdesc_t *));
+	*fn = fnbuf;
+
 	if (fn->fn_nargs != 0) {
 		debug(3, "die %llu: adding %d argument%s\n", off, fn->fn_nargs,
 		    (fn->fn_nargs > 1 ? "s" : ""));
 
-		fn->fn_args = xcalloc(sizeof (tdesc_t *) * fn->fn_nargs);
 		for (i = 0, arg = die_child(dw, die);
 		    arg != NULL && i < fn->fn_nargs;
 		    arg = die_sibling(dw, arg)) {
@@ -1280,8 +1399,9 @@ die_funcptr_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
  * that reflects said name.
  */
 static intr_t *
-die_base_name_parse(const char *name, char **newp)
+die_base_name_parse(atom_t *atom, atom_t **newp)
 {
+	const char *name = atom->value;
 	char buf[100];
 	char *base, *c;
 	int nlong = 0, nshort = 0, nchar = 0, nint = 0;
@@ -1350,7 +1470,7 @@ die_base_name_parse(const char *name, char **newp)
 	    (nlong > 1 ? "long " : ""),
 	    base);
 
-	*newp = xstrdup(buf);
+	*newp = atom_get(buf);
 	return (intr);
 }
 
@@ -1448,9 +1568,9 @@ die_base_create(dwarf_t *dw, Dwarf_Die base, Dwarf_Off off, tdesc_t *tdp)
 {
 	Dwarf_Unsigned sz;
 	intr_t *intr;
-	char *new;
+	atom_t *new;
 
-	debug(3, "die %llu: creating base type\n", off);
+	debug(3, "die %llu: creating base type %d\n", off, tdp->t_id);
 
 	/*
 	 * The compilers have their own clever (internally inconsistent) ideas
@@ -1463,16 +1583,14 @@ die_base_create(dwarf_t *dw, Dwarf_Die base, Dwarf_Off off, tdesc_t *tdp)
 	 */
 	(void) die_unsigned(dw, base, DW_AT_byte_size, &sz, DW_ATTR_REQ);
 
-	if (tdp->t_name == NULL)
+	if (tdp->t_name == ATOM_NULL)
 		terminate("die %llu: base type without name\n", off);
 
 	/* XXX make a name parser for float too */
 	if ((intr = die_base_name_parse(tdp->t_name, &new)) != NULL) {
 		/* Found it.  We'll use the parsed version */
-		debug(3, "die %llu: name \"%s\" remapped to \"%s\"\n", off,
-		    tdesc_name(tdp), new);
-
-		free(tdp->t_name);
+		debug(3, "die %llu: name \"%s\" remapped to \"%s\" size %d\n", off,
+		    tdesc_name(tdp), new->value, sz);
 		tdp->t_name = new;
 	} else {
 		/*
@@ -1502,15 +1620,17 @@ die_through_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp,
 
 	debug(3, "die %llu: creating %s\n", off, typename);
 
-	tdp->t_type = type;
-
 	if ((attr = die_attr(dw, die, DW_AT_type, 0)) != NULL) {
 		dwarf_dealloc(dw->dw_dw, attr, DW_DLA_ATTR);
 		tdp->t_tdesc = die_lookup_pass1(dw, die, DW_AT_type);
+		debug(3, "die %llu: to %llu\n", off, tdp->t_tdesc->t_id);
 	} else {
 		tdp->t_tdesc = tdesc_intr_void(dw);
 	}
 
+    /* set t_type after t_desc since a non-zero t_type requires t_desc */
+	tdp->t_type = type;
+
 	if (type == POINTER)
 		tdp->t_size = dw->dw_ptrsz;
 
@@ -1519,7 +1639,7 @@ die_through_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp,
 	if (type == TYPEDEF) {
 		iidesc_t *ii = xcalloc(sizeof (iidesc_t));
 		ii->ii_type = II_TYPE;
-		ii->ii_name = xstrdup(tdp->t_name);
+		ii->ii_name = tdp->t_name;
 		ii->ii_dtype = tdp;
 
 		iidesc_add(dw->dw_td->td_iihash, ii);
@@ -1556,6 +1676,50 @@ die_volatile_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 	die_through_create(dw, die, off, tdp, VOLATILE, "volatile");
 }
 
+static int
+die_isdiscriminated(dwarf_t *dw, Dwarf_Die die)
+{
+	Dwarf_Bool val;
+	return (die_bool(dw, die,
+	    DW_AT_APPLE_ptrauth_address_discriminated, &val, 0) && val);
+}
+
+/*ARGSUSED*/
+static void
+die_ptrauth_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
+{
+	Dwarf_Signed discriminator, key;
+	Dwarf_Attribute attr;
+	ptrauth_t *pta = xcalloc(sizeof (ptrauth_t));
+
+	debug(3, "die %llu: creating ptrauth\n", off);
+
+	if ((attr = die_attr(dw, die, DW_AT_type, 0)) == NULL) {
+		terminate("die %llu: no referenced type", die_off(dw, die));
+	}
+
+	dwarf_dealloc(dw->dw_dw, attr, DW_DLA_ATTR);
+
+	pta->pta_type = die_lookup_pass1(dw, die, DW_AT_type);
+
+	pta->pta_discriminated = die_isdiscriminated(dw, die);
+
+	die_signed(dw, die, DW_AT_APPLE_ptrauth_key, &key, 1);
+	pta->pta_key = key;
+	if (die_signed(dw, die, DW_AT_APPLE_ptrauth_extra_discriminator,
+	    &discriminator, 0)) {
+		pta->pta_discriminator = discriminator;
+	}
+
+	tdp->t_type = PTRAUTH;
+	tdp->t_size = dw->dw_ptrsz;
+	tdp->t_flags |= TDESC_F_RESOLVED;
+
+	tdp->t_ptrauth = pta;
+
+	debug(3, "die %llu: to %llu\n", off, pta->pta_type->t_id);
+}
+
 /*ARGSUSED3*/
 static void
 die_function_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
@@ -1563,7 +1727,7 @@ die_function_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 	Dwarf_Die arg;
 	Dwarf_Half tag;
 	iidesc_t *ii;
-	char *name;
+	atom_t *name;
 
 	debug(3, "die %llu: creating function definition\n", off);
 
@@ -1595,7 +1759,7 @@ die_function_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 		name = die_name(dw, die);
 	}
 	
-	if (name == NULL) {
+	if (name == ATOM_NULL) {
 		return; /* Subprogram without name. */
 	}
 #else
@@ -1612,9 +1776,9 @@ die_function_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 	ii->ii_type = die_isglobal(dw, die) ? II_GFUN : II_SFUN;
 	ii->ii_name = name;
 	if (ii->ii_type == II_SFUN)
-		ii->ii_owner = xstrdup(dw->dw_cuname);
+		ii->ii_owner = dw->dw_cuname;
 
-	debug(3, "die %llu: function %s is %s\n", off, ii->ii_name,
+	debug(3, "die %llu: function %s is %s\n", off, name,
 	    (ii->ii_type == II_GFUN ? "global" : "static"));
 
 	if (die_attr(dw, die, DW_AT_type, 0) != NULL)
@@ -1624,7 +1788,7 @@ die_function_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 
 	for (arg = die_child(dw, die); arg != NULL;
 	    arg = die_sibling(dw, arg)) {
-		char *name;
+		atom_t *tmp;
 
 		debug(3, "die %llu: looking at sub member at %llu\n",
 		    off, die_off(dw, die));
@@ -1637,12 +1801,12 @@ die_function_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 #endif /* __APPLE__ */
 
 #if !defined(__APPLE__)
-		if ((name = die_name(dw, arg)) == NULL) {
+		if ((tmp = die_name(dw, arg)) == ATOM_NULL) {
 			terminate("die %llu: func arg %d has no name\n",
 			    off, ii->ii_nargs + 1);
 		}
 #else
-		if ((name = die_name(dw, arg)) == NULL) {
+		if ((tmp = die_name(dw, arg)) == ATOM_NULL) {
 			/* C++ admits this, IOKit does it, we'll allow it. */
 			debug(1, "die %llu: func arg %d has no name\n",
 			    off, ii->ii_nargs + 1);
@@ -1651,8 +1815,7 @@ die_function_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 		}
 #endif /* __APPLE__ */
 
-		if (strcmp(name, "...") == 0) {
-			free(name);
+		if (strcmp(tmp->value, "...") == 0) {
 			ii->ii_vargs = 1;
 			continue;
 		}
@@ -1695,11 +1858,11 @@ static void
 die_variable_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 {
 	iidesc_t *ii;
-	char *name;
+	atom_t *name;
 
 	debug(3, "die %llu: creating object definition\n", off);
 
-	if (die_isdecl(dw, die) || (name = die_name(dw, die)) == NULL)
+	if (die_isdecl(dw, die) || (name = die_name(dw, die)) == ATOM_NULL)
 		return; /* skip prototypes and nameless objects */
 
 	ii = xcalloc(sizeof (iidesc_t));
@@ -1707,7 +1870,7 @@ die_variable_create(dwarf_t *dw, Dwarf_Die die, Dwarf_Off off, tdesc_t *tdp)
 	ii->ii_name = name;
 	ii->ii_dtype = die_lookup_pass1(dw, die, DW_AT_type);
 	if (ii->ii_type == II_SVAR)
-		ii->ii_owner = xstrdup(dw->dw_cuname);
+		ii->ii_owner = dw->dw_cuname;
 
 	iidesc_add(dw->dw_td->td_iihash, ii);
 }
@@ -1770,6 +1933,7 @@ static const die_creator_t die_creators[] = {
 	{ DW_TAG_variable,		DW_F_NOTDP,	die_variable_create },
 	{ DW_TAG_volatile_type,		0,		die_volatile_create },
 	{ DW_TAG_restrict_type,		0,		die_restrict_create },
+	{ DW_TAG_APPLE_ptrauth_type,	0,		die_ptrauth_create },
 	{ 0, NULL }
 };
 
@@ -1786,7 +1950,7 @@ die_tag2ctor(Dwarf_Half tag)
 	return (NULL);
 }
 
-static void
+static tdesc_t *
 die_create_one(dwarf_t *dw, Dwarf_Die die)
 {
 	Dwarf_Off off = die_off(dw, die);
@@ -1795,17 +1959,16 @@ die_create_one(dwarf_t *dw, Dwarf_Die die)
 	tdesc_t *tdp;
 
 	debug(3, "die %llu: create_one\n", off);
-
-	if (off > dw->dw_maxoff) {
-		terminate("illegal die offset %llu (max %llu)\n", off,
-		    dw->dw_maxoff);
+	if ((off < dw->dw_cuoff) || (off > dw->dw_maxoff)) {
+		debug(2, "die offset outside CU %llu (cuoff %llu max %llu)\n",
+				off, dw->dw_cuoff, dw->dw_maxoff);
 	}
 
 	tag = die_tag(dw, die);
 
 	if ((dc = die_tag2ctor(tag)) == NULL) {
 		debug(2, "die %llu: ignoring tag type %x\n", off, tag);
-		return;
+		return (NULL);
 	}
 
 	if ((tdp = tdesc_lookup(dw, off)) == NULL &&
@@ -1821,11 +1984,13 @@ die_create_one(dwarf_t *dw, Dwarf_Die die)
 #if defined(__APPLE__)
 #warning BOGUS workaround (nameless base die emitted by gcc)!
 	/* gcc 5402 emits nameless die that are base types. Workaround imminent failure. */
-	if (tdp != NULL && NULL == tdp->t_name && dc->dc_create == die_base_create)
-		tdp->t_name = xstrdup("unsigned int");
+	if (tdp != NULL && ATOM_NULL == tdp->t_name && dc->dc_create == die_base_create)
+		tdp->t_name = atom_get("unsigned int");
 #endif /* __APPLE__ */
 
 	dc->dc_create(dw, die, off, tdp);
+
+	return (tdp);
 }
 
 static void
@@ -1851,6 +2016,7 @@ static tdtrav_cb_f die_resolvers[] = {
 	NULL,			/* volatile */
 	NULL,			/* const */
 	NULL,			/* restrict */
+	NULL,			/* ptrauth */
 };
 
 static tdtrav_cb_f die_fail_reporters[] = {
@@ -1868,6 +2034,7 @@ static tdtrav_cb_f die_fail_reporters[] = {
 	NULL,			/* volatile */
 	NULL,			/* const */
 	NULL,			/* restrict */
+	NULL,			/* ptrauth */
 };
 
 static void
@@ -1903,90 +2070,206 @@ die_resolve(dwarf_t *dw)
 
 /*ARGSUSED*/
 int
-dw_read(tdata_t *td, Elf *elf, const char *filename)
+dw_read(Elf *elf, const char *filename, const char *unitmatch, int verbose, tdata_t **mstrtd)
 {
+	dwarf_t _dw;
+	dwarf_t *dw = &_dw;
 	Dwarf_Unsigned abboff, hdrlen, nxthdr;
 	Dwarf_Half vers, addrsz;
-	Dwarf_Die cu, child;
-	dwarf_t dw;
-	char *prod = NULL;
+	Dwarf_Die child;
+	char *prod;
 	int rc;
+	int cucount = 0;
+	int cumerged = 0;
+	cu_data_t *cufirst = NULL;
+	cu_data_t *culast = NULL;
+	cu_data_t *cunext = NULL;
+	cu_data_t *cudata;
+	tdata_t *cutd = NULL;
+	merge_cb_data_t mcd = {0};
 
-	bzero(&dw, sizeof (dwarf_t));
-	dw.dw_td = td;
-	dw.dw_ptrsz = elf_ptrsz(elf);
-	dw.dw_mfgtid_last = TID_MFGTID_BASE;
-	dw.dw_tidhash = hash_new(TDESC_HASH_BUCKETS, tdesc_idhash, tdesc_idcmp);
-	dw.dw_fwdhash = hash_new(TDESC_HASH_BUCKETS, tdesc_namehash,
-	    tdesc_namecmp);
-	dw.dw_enumhash = hash_new(TDESC_HASH_BUCKETS, tdesc_namehash,
-	    tdesc_namecmp);
+	bzero(dw, sizeof (dwarf_t));
+	dw->dw_ptrsz = elf_ptrsz(elf);
 
-	if ((rc = dwarf_elf_init(elf, DW_DLC_READ, NULL, NULL, &dw.dw_dw,
-	    &dw.dw_err)) == DW_DLV_NO_ENTRY) {
-		errno = ENOENT;
+	if ((rc = dwarf_elf_init(elf, DW_DLC_READ, NULL, NULL, &dw->dw_dw,
+		&dw->dw_err)) == DW_DLV_NO_ENTRY) {
 		return (-1);
 	} else if (rc != DW_DLV_OK) {
-		if (dwarf_errno(dw.dw_err) == DW_DLE_DEBUG_INFO_NULL) {
+		if (dwarf_errno(dw->dw_err) == DW_DLE_DEBUG_INFO_NULL) {
 			/*
 			 * There's no type data in the DWARF section, but
 			 * libdwarf is too clever to handle that properly.
 			 */
-			return (0);
+			return (-1);
 		}
 
 		terminate("failed to initialize DWARF: %s\n",
-		    dwarf_errmsg(dw.dw_err));
+			dwarf_errmsg(dw->dw_err));
 	}
 
-	if ((rc = dwarf_next_cu_header(dw.dw_dw, &hdrlen, &vers, &abboff,
-	    &addrsz, &nxthdr, &dw.dw_err)) != DW_DLV_OK ||
-	    (cu = die_sibling(&dw, NULL)) == NULL ||
-	    (child = die_child(&dw, cu)) == NULL)
-		terminate("file does not contain dwarf type data "
-		    "(try compiling with -g)\n");
+	/* First pass to find the compilation units we will process */
 
-	dw.dw_maxoff = nxthdr - 1;
+	while (1) {
+		dw->dw_cuname = ATOM_NULL;
+		if ((rc = dwarf_next_cu_header(dw->dw_dw, &hdrlen, &vers, &abboff,
+			&addrsz, &nxthdr, &dw->dw_err)) != DW_DLV_OK) {
+			break;
+		}
 
-	if (dw.dw_maxoff > TID_FILEMAX)
-		terminate("file contains too many types\n");
+		dw->dw_maxoff = nxthdr - 1;
+		if (dw->dw_maxoff > TID_FILEMAX)
+			terminate("file contains too many types\n");
 
-	debug(1, "DWARF version: %d\n", vers);
-	if (vers != DWARF_VERSION) {
-		terminate("file contains incompatible version %d DWARF code "
-		    "(version 2 required)\n", vers);
+		debug(4, "DWARF version: %d\n", vers);
+		if (vers != DWARF_VERSION && vers != DWARF_VERSION3 && vers != DWARF_VERSION4) {
+			terminate("file contains incompatible version %d DWARF code "
+				"(version 2/3/4 required)\n", vers);
+		}
+
+		Dwarf_Error error = NULL;
+		rc = dwarf_siblingof(dw->dw_dw, NULL, &dw->dw_cu, &error);
+		if(rc != DW_DLV_OK) {
+			debug(4, "Error in dwarf_siblingof on CU die (%s)\n", dwarf_errmsg(error));
+			continue;
+		}
+
+		if (die_string(dw, dw->dw_cu, DW_AT_producer, &prod, 0)) {
+			debug(4, "DWARF emitter: %s\n", prod);
+			free(prod);
+		}
+
+		dw->dw_cuname = die_name(dw, dw->dw_cu);
+		if (!dw->dw_cuname) {
+			debug(1, "no name for CU\n");
+			continue;
+		}
+
+		if (unitmatch) {
+			/* Only extract from CUs built from a directory path
+			 * containing the string unitmatch. */
+			char *comp_dir;
+			if (!die_string(dw, dw->dw_cu, DW_AT_comp_dir, &comp_dir, 0)) {
+				debug(1, "no AT_comp_dir for CU\n");
+				continue;
+			}
+			int skip = (!strstr(comp_dir, unitmatch));
+			free(comp_dir);
+			if (skip) {
+				continue;
+			}
+		}
+
+		Dwarf_Unsigned lang;
+		rc = dwarf_srclang(dw->dw_cu, &lang, &error);
+		if (rc == DW_DLV_OK) {
+			/* This code doesn't handle the DW_TAG_class_type,
+			 *  or scope c++ definitions properly.
+			 */
+			if (lang == DW_LANG_C_plus_plus
+			    || lang == DW_LANG_ObjC_plus_plus
+			    || lang == DW_LANG_C_plus_plus_03
+			    || lang == DW_LANG_C_plus_plus_11
+			    || lang == DW_LANG_C_plus_plus_14) {
+				continue;
+			}
+		}
+
+		cudata = xcalloc(sizeof(*cudata));
+		if (!cufirst) {
+			cufirst = cudata;
+		}
+		if (culast) {
+			culast->cud_next = cudata;
+		}
+		culast = cudata;
+		cucount++;
+		cudata->cud_name = dw->dw_cuname;
+		cudata->cud_nxthdr = nxthdr;
 	}
 
-	if (die_string(&dw, cu, DW_AT_producer, &prod, 0)) {
-		debug(1, "DWARF emitter: %s\n", prod);
-		free(prod);
-	}
+	ctfmerge_prepare(cucount);
 
-	if ((dw.dw_cuname = die_name(&dw, cu)) != NULL) {
-		char *base = xstrdup(basename(dw.dw_cuname));
-		free(dw.dw_cuname);
-		dw.dw_cuname = base;
+	/* Second pass to extract & merge */
 
-		debug(1, "CU name: %s\n", dw.dw_cuname);
-	}
+	dw->dw_tidhash = alist_new(TDESC_HASH_BUCKETS);
 
-	die_create(&dw, child);
+	for (cudata = cunext = cufirst; cudata; cudata = cunext) {
+		dw->dw_cuname = ATOM_NULL;
+		if ((rc = dwarf_next_cu_header(dw->dw_dw, &hdrlen, &vers, &abboff,
+			&addrsz, &nxthdr, &dw->dw_err)) != DW_DLV_OK) {
+			break;
+		}
+		if (nxthdr != cudata->cud_nxthdr) {
+			continue;
+		}
+		cunext = cudata->cud_next;
 
-	if ((rc = dwarf_next_cu_header(dw.dw_dw, &hdrlen, &vers, &abboff,
-	    &addrsz, &nxthdr, &dw.dw_err)) != DW_DLV_NO_ENTRY)
-		terminate("multiple compilation units not supported\n");
+		Dwarf_Die no_die = 0;
+		Dwarf_Error error = NULL;
+		rc = dwarf_siblingof(dw->dw_dw, no_die, &dw->dw_cu, &error);
+		if(rc != DW_DLV_OK) {
+			terminate("Error in dwarf_siblingof on CU die (%s)\n", dwarf_errmsg(error));
+		}
 
-	(void) dwarf_finish(dw.dw_dw, &dw.dw_err);
+		dw->dw_cuoff = die_off(dw, dw->dw_cu);
+		dw->dw_cuname = cudata->cud_name;
+		if (verbose && (cucount > 1)) {
+			printf("[%.02fM] %s\n", nxthdr / 1024. / 1024., dw->dw_cuname->value);
+		}
+		char *tmp = xstrdup(dw->dw_cuname->value);
+		dw->dw_cuname = atom_get(basename(tmp));
+		free(tmp);
 
-	die_resolve(&dw);
+		rc = dwarf_child(dw->dw_cu, &child, &error);
+		if (rc != DW_DLV_OK) {
+			printf("Error in dwarf_child on CU die \n");
+			continue;
+		}
+
+		dw->dw_mfgtid_last = TID_MFGTID_BASE;
+		dw->dw_fwdhash = hash_new(TDESC_HASH_BUCKETS, tdesc_namehash,
+		    tdesc_namecmp);
+		dw->dw_enumhash = hash_new(TDESC_HASH_BUCKETS, tdesc_namehash,
+		    tdesc_namecmp);
+
+		dw->dw_void = NULL;
+		dw->dw_long = NULL;
+		dw->dw_maxoff = nxthdr - 1;
+
+		dw->dw_td = tdata_new();
+		die_create(dw, child);
+		die_resolve(dw);
 
 #if !defined(__APPLE__)
-	cvt_fixups(td, dw.dw_ptrsz);
+		cvt_fixups(td, dw->dw_ptrsz);
 #else
-	/* Ignore Solaris gore. See on-src-20080707/usr/src/tools/ctf/cvt/fixup_tdescs.c */
+		/* Ignore Solaris gore. See on-src-20080707/usr/src/tools/ctf/cvt/fixup_tdescs.c */
 #endif
 
-	/* leak the dwarf_t */
+		cutd = tdata_new();
+		debug(1, "mergeto %s to cutd\n", dw->dw_cuname->value);
+		merge_into_master(&mcd, dw->dw_td, cutd, NULL, 1);
 
-	return (0);
+		ctfmerge_add_td(cutd, dw->dw_cuname->value);
+
+		alist_clear(dw->dw_tidhash);
+		hash_free(dw->dw_fwdhash,  NULL, NULL);
+		hash_free(dw->dw_enumhash, NULL, NULL);
+		tdata_free(dw->dw_td);
+
+		cumerged++;
+	}
+
+	alist_free(dw->dw_tidhash);
+	merge_cb_data_destroy(&mcd);
+
+	debug(1, "dwarf_finish\n");
+	(void) dwarf_finish(dw->dw_dw, &dw->dw_err);
+
+	if (cumerged) {
+		*mstrtd = ctfmerge_done();
+		return (0);
+	} else {
+		return (-1);
+	}
 }
